"""Representation of a Tractive trackable object (pet)."""

from typing import Any

from .data_object import DataObject


class TrackableObject(DataObject):
    """Representation of a Tractive trackable object (pet)."""

    async def details(self) -> dict[str, Any]:
        """Get trackable object details."""
        trackable_object: dict[str, Any] = await self._api.request(
            f"trackable_object/{self._id}"
        )
        return trackable_object

    async def health_overview(self) -> dict[str, Any]:
        """Get health overview data including activity, sleep, rest, and health metrics.

        Returns health_overview data from the APS API endpoint.
        Replaces the deprecated wellness_overview message.
        """
        health_overview: dict[str, Any] = await self._api.request(
            f"pet/{self._id}/health/overview",
            base_url=self._api.APS_API_URL,
        )
        return health_overview



"""Base class for Tractive data objects."""

from __future__ import annotations

from typing import Any

from .api import API


class DataObject:
    """Base class for Tractive data objects."""

    def __init__(self, api: API, data: dict[str, Any]) -> None:
        """Initialize the data object."""
        self._api = api
        self._id: str = data["_id"]
        self.type: str = data["_type"]

    def __repr__(self) -> str:
        """Return string representation of the data object."""
        return f"<{self.__class__.__name__} id={self._id} type={self.type}>"



"""Tests for channel module."""

from __future__ import annotations

import asyncio
from asyncio.exceptions import TimeoutError as AIOTimeoutError
from http import HTTPStatus
from typing import Any, cast
from unittest.mock import AsyncMock, MagicMock

import pytest
from aiohttp import ClientResponse
from aiohttp.client_exceptions import ClientResponseError

from aiotractive.channel import Channel
from aiotractive.exceptions import TractiveError, UnauthorizedError


@pytest.fixture
def mock_api() -> MagicMock:
    """Create a mock API instance."""
    api = MagicMock()
    api.auth_headers = AsyncMock(return_value={"Authorization": "Bearer test"})
    api.retry_count = 3
    api.retry_delay = MagicMock(return_value=0.1)

    return api


@pytest.fixture
def channel(mock_api: MagicMock) -> Channel:
    """Create a Channel instance with mocked API."""
    return Channel(mock_api)


def create_mock_response(events: list[bytes]) -> MagicMock:
    """Create a mock response with async iterator over events."""
    response = MagicMock(spec=ClientResponse)
    response.content = AsyncIterator(events)
    response.status = 200
    response.request_info = MagicMock()
    response.history = ()
    response.reason = "OK"
    return response


class AsyncIterator:
    """Async iterator for mocking response.content."""

    def __init__(self, items: list[bytes]) -> None:
        """Initialize."""
        self.items = iter(items)
        self._exhausted = asyncio.Event()

    def __aiter__(self) -> AsyncIterator:
        """Async iterator enter."""
        return self

    async def __anext__(self) -> bytes:
        """Get next item asynchronously."""
        try:
            return next(self.items)
        except StopIteration:
            # Block indefinitely until cancelled, simulating waiting for more data
            await self._exhausted.wait()
            raise StopAsyncIteration from None


async def test_listen_receives_event(channel: Channel, mock_api: MagicMock) -> None:
    """Test that _listen puts valid events into the queue."""
    event_data = b'{"message": "test_event", "data": {"id": "123"}}'
    mock_response = create_mock_response([event_data])

    mock_context = AsyncMock()
    mock_context.__aenter__.return_value = mock_response
    mock_api.session.request.return_value = mock_context

    task = asyncio.create_task(channel._listen())
    await asyncio.sleep(0.1)
    task.cancel()
    await task

    events = []
    while not channel._queue.empty():
        events.append(await channel._queue.get())

    assert events
    assert len(events) == 2  # One event + one cancelled event
    event = events[0]
    assert event["event"]["message"] == "test_event"
    assert event["event"]["data"] == {"id": "123"}


async def test_listen_multiple_events(channel: Channel, mock_api: MagicMock) -> None:
    """Test that _listen processes multiple events in sequence."""
    input_events = [
        b'{"message": "keep-alive"}',
        b'{"message": "tracker_status", "id": "1"}',
        b'{"message": "handshake"}',
        b'{"message": "position", "id": "2"}',
    ]
    mock_response = create_mock_response(input_events)

    mock_context = AsyncMock()
    mock_context.__aenter__.return_value = mock_response
    mock_api.session.request.return_value = mock_context

    task = asyncio.create_task(channel._listen())
    await asyncio.sleep(0.1)
    task.cancel()
    await task

    assert channel._last_keep_alive is not None

    events = []
    while not channel._queue.empty():
        events.append(await channel._queue.get())

    assert len(events) == 3  # Two events + one cancelled event
    assert events[0]["event"]["message"] == "tracker_status"
    assert events[1]["event"]["message"] == "position"


async def test_listen_keep_alive(channel: Channel, mock_api: MagicMock) -> None:
    """Test that _listen updates _last_keep_alive on keep-alive message."""
    event_data = b'{"message": "keep-alive"}'
    mock_response = create_mock_response([event_data])

    mock_context = AsyncMock()
    mock_context.__aenter__.return_value = mock_response
    mock_api.session.request.return_value = mock_context

    assert channel._last_keep_alive is None

    task = asyncio.create_task(channel._listen())
    await asyncio.sleep(0.1)
    task.cancel()
    await task

    assert channel._last_keep_alive is not None

    events = []
    while not channel._queue.empty():
        events.append(await channel._queue.get())

    # Queue should only have the cancelled event, no keep-alive events
    assert events
    assert len(events) == 1
    event = events[0]
    assert event["type"] == "cancelled"


async def test_listen_handshake(channel: Channel, mock_api: MagicMock) -> None:
    """Test that _listen ignores handshake messages."""
    event_data = b'{"message": "handshake"}'
    mock_response = create_mock_response([event_data])

    mock_context = AsyncMock()
    mock_context.__aenter__.return_value = mock_response
    mock_api.session.request.return_value = mock_context

    task = asyncio.create_task(channel._listen())
    await asyncio.sleep(0.1)
    task.cancel()
    await task

    events = []
    while not channel._queue.empty():
        events.append(await channel._queue.get())

    # Queue should only have the cancelled event, no keep-alive events
    assert events
    assert len(events) == 1
    event = events[0]
    assert event["type"] == "cancelled"


async def test_listen_timeout(channel: Channel, mock_api: MagicMock) -> None:
    """Test that _listen continues loop on AIOTimeoutError."""
    call_count = 0
    mock_api.retry_delay.return_value = 0

    def side_effect(*args: Any, **kwargs: Any) -> AsyncMock:  # noqa: ANN401,ARG001
        nonlocal call_count
        call_count += 1
        if call_count == 1:
            raise AIOTimeoutError
        mock_context = AsyncMock()
        mock_context.__aenter__.return_value = create_mock_response(
            [b'{"message": "tracker_status"}']
        )
        return mock_context

    mock_api.session.request.side_effect = side_effect

    task = asyncio.create_task(channel._listen())
    await asyncio.sleep(0.1)
    task.cancel()
    await task

    assert call_count >= 2


async def test_listen_unauthorized_401(channel: Channel, mock_api: MagicMock) -> None:
    """Test that _listen handles 401 ClientResponseError."""
    exc = ClientResponseError(
        request_info=MagicMock(),
        history=(),
        status=HTTPStatus.UNAUTHORIZED,
        message="Unauthorized",
    )
    mock_api.session.request.side_effect = exc

    task = asyncio.create_task(channel._listen())
    await asyncio.sleep(0.1)
    task.cancel()

    events = []
    while not channel._queue.empty():
        events.append(await channel._queue.get())

    assert events
    assert len(events) == 1
    event = events[0]
    assert event["type"] == "error"
    assert isinstance(event["error"], UnauthorizedError)
    assert event["error"].__cause__ is exc


async def test_listen_unauthorized_404(channel: Channel, mock_api: MagicMock) -> None:
    """Test that _listen handles 404 ClientResponseError."""
    exc = ClientResponseError(
        request_info=MagicMock(),
        history=(),
        status=HTTPStatus.NOT_FOUND,
        message="Not found",
    )
    mock_api.session.request.side_effect = exc

    task = asyncio.create_task(channel._listen())
    await asyncio.sleep(0.1)
    task.cancel()

    events = []
    while not channel._queue.empty():
        events.append(await channel._queue.get())

    assert events
    assert len(events) == 1
    event = events[0]
    assert event["type"] == "error"
    assert isinstance(event["error"], TractiveError)
    assert event["error"].__cause__ is exc


async def test_listen_exception(channel: Channel, mock_api: MagicMock) -> None:
    """Test that _listen handles generic exceptions."""
    exc = ValueError("Something went wrong")
    mock_api.retry_delay.return_value = 0
    mock_api.session.request.side_effect = exc

    task = asyncio.create_task(channel._listen())
    await asyncio.sleep(0.1)
    task.cancel()

    events = []
    while not channel._queue.empty():
        events.append(await channel._queue.get())

    assert events
    assert len(events) == 1
    event = events[0]
    assert event["type"] == "error"
    assert isinstance(event["error"], TractiveError)
    assert event["error"].__cause__ is exc


async def test_listen_retry_on_429_client_error(
    channel: Channel, mock_api: MagicMock
) -> None:
    """Test that _listen retries on 429 ClientResponseError."""
    call_count = 0

    def side_effect(*args: Any, **kwargs: Any) -> AsyncMock:  # noqa: ANN401,ARG001
        nonlocal call_count
        call_count += 1
        if call_count == 1:
            raise ClientResponseError(
                request_info=MagicMock(),
                history=(),
                status=HTTPStatus.TOO_MANY_REQUESTS,
                message="Too Many Requests",
            )
        mock_context = AsyncMock()
        mock_context.__aenter__.return_value = create_mock_response(
            [b'{"message": "tracker_status"}']
        )
        return mock_context

    mock_api.session.request.side_effect = side_effect

    task = asyncio.create_task(channel._listen())
    await asyncio.sleep(0.2)
    task.cancel()
    await task

    assert call_count >= 2
    assert cast(MagicMock, channel._retry_delay).called


async def test_listen_retry_exhausted_on_429(
    channel: Channel, mock_api: MagicMock
) -> None:
    """Test that _listen gives up after exhausting retries on 429."""
    mock_api.retry_count = 1

    exc = ClientResponseError(
        request_info=MagicMock(),
        history=(),
        status=HTTPStatus.TOO_MANY_REQUESTS,
        message="Too Many Requests",
    )
    mock_api.session.request.side_effect = exc

    task = asyncio.create_task(channel._listen())
    await asyncio.sleep(0.5)
    task.cancel()

    events = []
    while not channel._queue.empty():
        events.append(await channel._queue.get())

    assert events
    assert len(events) == 1
    event = events[0]
    assert event["type"] == "error"
    assert isinstance(event["error"], TractiveError)
    assert "Too Many Requests" in str(event["error"])



"""Representation of a Tractive tracker device."""

from typing import Any, ClassVar

from .data_object import DataObject


class Tracker(DataObject):
    """Representation of a Tractive tracker device."""

    ACTIONS: ClassVar[dict[bool, str]] = {True: "on", False: "off"}

    async def details(self) -> dict[str, Any]:
        """Get tracker details."""
        details: dict[str, Any] = await self._api.request(f"tracker/{self._id}")
        return details

    async def hw_info(self) -> dict[str, Any]:
        """Get hardware info for the tracker."""
        hw_info: dict[str, Any] = await self._api.request(
            f"device_hw_report/{self._id}/"
        )
        return hw_info

    async def pos_report(self) -> dict[str, Any]:
        """Get position report for the tracker."""
        pos_report: dict[str, Any] = await self._api.request(
            f"device_pos_report/{self._id}"
        )
        return pos_report

    async def positions(
        self, time_from: float, time_to: float, fmt: str
    ) -> dict[str, Any]:
        """Get positions for the tracker within a time range."""
        url = f"tracker/{self._id}/positions"
        params = {
            "time_from": time_from,
            "time_to": time_to,
            "format": fmt,
        }
        positions: dict[str, Any] = await self._api.request(url, params=params)
        return positions

    async def set_buzzer_active(self, active: bool) -> dict[str, Any]:
        """Enable or disable the buzzer on the tracker."""
        action = self.ACTIONS[active]
        result: dict[str, Any] = await self._api.request(
            f"tracker/{self._id}/command/buzzer_control/{action}"
        )
        return result

    async def set_led_active(self, active: bool) -> dict[str, Any]:
        """Enable or disable the LED on the tracker."""
        action = self.ACTIONS[active]
        result: dict[str, Any] = await self._api.request(
            f"tracker/{self._id}/command/led_control/{action}"
        )
        return result

    async def set_live_tracking_active(self, active: bool) -> dict[str, Any]:
        """Enable or disable live tracking mode on the tracker."""
        action = self.ACTIONS[active]
        result: dict[str, Any] = await self._api.request(
            f"tracker/{self._id}/command/live_tracking/{action}"
        )
        return result



"""Entrypoint for the Tractive REST API."""

from __future__ import annotations

from collections.abc import AsyncIterator
from types import TracebackType
from typing import Any

from .api import API
from .channel import Channel
from .trackable_object import TrackableObject
from .tracker import Tracker


class Tractive:
    """Asynchronous Python client for the Tractive REST API."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize the client."""
        self._api = API(*args, **kwargs)

    async def authenticate(self) -> dict[str, Any] | None:
        """Authenticate the client."""
        return await self._api.authenticate()

    async def trackers(self) -> list[Tracker]:
        """Get all trackers for the authenticated user."""
        trackers: list[dict[str, Any]] = await self._api.request(
            f"user/{await self._api.user_id()}/trackers"
        )
        return [Tracker(self._api, t) for t in trackers]

    def tracker(self, tracker_id: str) -> Tracker:
        """Get tracker by ID."""
        return Tracker(self._api, {"_id": tracker_id, "_type": "tracker"})

    def trackable_object(self, trackable_id: str) -> TrackableObject:
        """Get trackable object by ID."""
        return TrackableObject(self._api, {"_id": trackable_id, "_type": "pet"})

    async def trackable_objects(self) -> list[TrackableObject]:
        """Get all trackable objects for the authenticated user."""
        trackable_objects: list[dict[str, Any]] = await self._api.request(
            f"user/{await self._api.user_id()}/trackable_objects"
        )
        return [TrackableObject(self._api, t) for t in trackable_objects]

    async def events(self) -> AsyncIterator[dict[str, Any]]:
        """Listen for real-time events from the Tractive API."""
        async for event in Channel(self._api).listen():
            yield event

    async def close(self) -> None:
        """Close open client session."""
        await self._api.close()

    async def __aenter__(self) -> Tractive:  # noqa: PYI034
        """Async enter."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Async exit."""
        await self.close()



"""Tests for the API client."""

from __future__ import annotations

from typing import Any

from aioresponses import aioresponses

from aiotractive import Tractive
from aiotractive.api import API

API_URL = str(API.API_URL)


def mock_auth(mock: aioresponses, auth_response: dict[str, Any]) -> None:
    """Add authentication endpoint mock to aioresponses."""
    mock.post(f"{API_URL}auth/token", payload=auth_response)


async def test_auth_success(auth_response: dict[str, Any]) -> None:
    """Test successful authentication returns user_id and access_token."""
    with aioresponses() as mock:
        mock_auth(mock, auth_response)

        async with Tractive("test@example.com", "password") as client:
            creds = await client.authenticate()

    assert creds is not None

    assert creds["user_id"] == "test_user_123"
    assert creds["access_token"] == "test_access_token_xyz"  # noqa: S105



"""Channel for real-time events from the Tractive REST API."""

from __future__ import annotations

import asyncio
import time
from asyncio.exceptions import TimeoutError as AIOTimeoutError
from collections.abc import AsyncIterator
from http import HTTPStatus
from typing import TYPE_CHECKING, Any

import aiohttp
import orjson
from aiohttp.client_exceptions import ClientResponseError

from .api import API
from .exceptions import DisconnectedError, TractiveError, UnauthorizedError


class Channel:
    """Channel for real-time events from the Tractive REST API."""

    CHANNEL_URL = "https://channel.tractive.com/3/channel"
    IGNORE_MESSAGES = ("handshake", "keep-alive")

    KEEP_ALIVE_TIMEOUT = 60  # seconds
    CHECK_CONNECTION_TIME = 5  # seconds

    def __init__(self, api: API) -> None:
        """Initialize the channel."""
        self._api = api
        self._last_keep_alive: float | None = None
        self._listen_task: asyncio.Task[None] | None = None
        self._check_connection_task: asyncio.Task[None] | None = None
        self._queue: asyncio.Queue[dict[str, Any]] = asyncio.Queue()

        self._retry_count = api.retry_count
        self._retry_delay = api.retry_delay

    async def listen(self) -> AsyncIterator[dict[str, Any]]:
        """Listen for real-time events from the Tractive API."""
        self._check_connection_task = asyncio.create_task(self._check_connection())
        self._listen_task = asyncio.create_task(self._listen())
        while True:
            event = await self._queue.get()
            self._queue.task_done()

            if event["type"] == "event":
                yield event["event"]

            if event["type"] == "error":
                self._check_connection_task.cancel()
                await self._check_connection_task

                self._listen_task.cancel()
                await self._listen_task

                raise event["error"]

            if event["type"] == "cancelled":
                self._listen_task.cancel()

                await self._listen_task
                raise DisconnectedError from event["error"]

    async def _listen(self) -> None:
        if TYPE_CHECKING:
            assert self._api.session is not None

        exc: TractiveError

        attempt = 0
        while True:
            attempt += 1
            try:
                async with self._api.session.request(
                    "POST",
                    self.CHANNEL_URL,
                    headers=await self._api.auth_headers(),
                    timeout=aiohttp.ClientTimeout(
                        total=None,
                        connect=10,
                        sock_connect=10,
                        sock_read=None,
                        ceil_threshold=5,
                    ),
                ) as response:
                    response.raise_for_status()
                    async for data in response.content:
                        event = orjson.loads(data)
                        if event["message"] == "keep-alive":
                            self._last_keep_alive = time.time()
                            continue
                        if event["message"] in self.IGNORE_MESSAGES:
                            continue
                        await self._queue.put({"type": "event", "event": event})
            except AIOTimeoutError:
                continue
            except ClientResponseError as error:
                if error.status in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
                    # Non-recoverable failures
                    exc = UnauthorizedError(str(error))
                    exc.__cause__ = error
                    await self._queue.put({"type": "error", "error": exc})
                    return

                if (
                    error.status == HTTPStatus.TOO_MANY_REQUESTS
                    or error.status >= HTTPStatus.INTERNAL_SERVER_ERROR
                ):
                    # Recoverable failures
                    if attempt <= self._retry_count:
                        delay = self._retry_delay(attempt)
                        await asyncio.sleep(delay)
                        continue

                    exc = TractiveError(str(error))
                    exc.__cause__ = error
                    await self._queue.put({"type": "error", "error": exc})
                    return

                # Other client errors, treat as non-recoverable
                exc = TractiveError(str(error))
                exc.__cause__ = error
                await self._queue.put({"type": "error", "error": exc})
                return

            except asyncio.CancelledError as error:
                await self._queue.put({"type": "cancelled", "error": error})
                return

            except Exception as error:  # noqa: BLE001
                exc = TractiveError(str(error))
                exc.__cause__ = error
                await self._queue.put({"type": "error", "error": exc})
                return

    async def _check_connection(self) -> None:
        try:
            while True:
                if self._last_keep_alive is not None and (
                    time.time() - self._last_keep_alive > self.KEEP_ALIVE_TIMEOUT
                ):
                    if self._listen_task is not None:
                        self._listen_task.cancel()
                    return

                await asyncio.sleep(self.CHECK_CONNECTION_TIME)
        except asyncio.CancelledError:
            return



"""Low level client for the Tractive REST API."""

from __future__ import annotations

import asyncio
import logging
import random
import time
from collections.abc import Callable
from http import HTTPStatus
from typing import TYPE_CHECKING, Any

import aiohttp
from aiohttp.client_exceptions import ClientResponseError
from yarl import URL

from .exceptions import NotFoundError, TractiveError, UnauthorizedError

CLIENT_ID = "625e533dc3c3b41c28a669f0"

_LOGGER = logging.getLogger(__name__)


class API:
    """Client for the API handling auth, requests, retries, and error mapping."""

    API_URL = URL("https://graph.tractive.com/4/")
    APS_API_URL = URL("https://aps-api.tractive.com/api/1/")

    DEFAULT_TIMEOUT = 10

    TOKEN_URI = "auth/token"  # noqa: S105

    def __init__(
        self,
        login: str,
        password: str,
        client_id: str = CLIENT_ID,
        timeout: int = DEFAULT_TIMEOUT,
        loop: asyncio.AbstractEventLoop | None = None,
        session: aiohttp.ClientSession | None = None,
        retry_count: int = 3,
        retry_delay: Callable[[int], float] = lambda attempt: 4**attempt
        + random.uniform(0, 3),  # noqa: S311
    ) -> None:
        """Initialize."""
        self._login = login
        self._password = password
        self._client_id = client_id
        self._timeout = timeout

        self.session: aiohttp.ClientSession | None = session
        self._close_session: bool = False

        if self.session is None:
            loop = loop or asyncio.get_event_loop()
            self.session = aiohttp.ClientSession(raise_for_status=True)
            self._close_session = True

        self._user_credentials: dict[str, Any] | None = None
        self._auth_headers: dict[str, str] | None = None

        self.retry_count = retry_count
        self.retry_delay = retry_delay

    async def user_id(self) -> str:
        """Get user ID."""
        await self.authenticate()
        if TYPE_CHECKING:
            assert self._user_credentials is not None
        user_id: str = self._user_credentials["user_id"]
        return user_id

    async def auth_headers(self) -> dict[str, str]:
        """Get authentication headers."""
        await self.authenticate()
        if TYPE_CHECKING:
            assert self._auth_headers is not None
        return {**self.base_headers(), **self._auth_headers}

    async def request(self, *args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
        """Perform request with error wrapping."""
        try:
            return await self.raw_request(*args, **kwargs)
        except ClientResponseError as error:
            if error.status in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
                raise UnauthorizedError from error
            if error.status == HTTPStatus.NOT_FOUND:
                raise NotFoundError from error
            raise TractiveError from error
        except Exception as error:
            raise TractiveError from error

    async def raw_request(
        self,
        uri: str,
        params: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
        method: str = "GET",
        attempt: int = 1,
        base_url: URL = API_URL,
    ) -> Any:  # noqa: ANN401
        """Perform request."""
        if TYPE_CHECKING:
            assert self.session is not None
        async with self.session.request(
            method,
            base_url.join(URL(uri)).update_query(params),
            json=data,
            headers=await self.auth_headers(),
            timeout=aiohttp.ClientTimeout(total=self._timeout),
        ) as response:
            _LOGGER.debug("Request %s, status: %s", response.url, response.status)

            if response.status == HTTPStatus.TOO_MANY_REQUESTS:
                if attempt <= self.retry_count:
                    delay = self.retry_delay(attempt)
                    _LOGGER.info("Request limit exceeded, retrying in %s second", delay)
                    await asyncio.sleep(delay)
                    return await self.raw_request(
                        uri,
                        params,
                        data,
                        method,
                        attempt=attempt + 1,
                        base_url=base_url,
                    )
                raise TractiveError("Request limit exceeded")

            if (
                "Content-Type" in response.headers
                and "application/json" in response.headers["Content-Type"]
            ):
                return await response.json()
            return await response.read()

    async def authenticate(self) -> dict[str, Any] | None:
        """Perform authentication."""
        if (
            self._user_credentials is not None
            and self._user_credentials["expires_at"] - time.time() < 3600  # noqa: PLR2004
        ):
            self._user_credentials = None
            self._auth_headers = None

        if self._user_credentials is not None:
            return self._user_credentials

        if TYPE_CHECKING:
            assert self.session is not None
        try:
            async with self.session.request(
                "POST",
                self.API_URL.join(URL(self.TOKEN_URI)),
                json={
                    "platform_email": self._login,
                    "platform_token": self._password,
                    "grant_type": "tractive",
                },
                headers=self.base_headers(),
                timeout=aiohttp.ClientTimeout(total=self._timeout),
            ) as response:
                if (
                    "Content-Type" in response.headers
                    and "application/json" in response.headers["Content-Type"]
                ):
                    self._user_credentials = await response.json()
                    self._auth_headers = {
                        "x-tractive-user": self._user_credentials["user_id"],
                        "authorization": (
                            f"Bearer {self._user_credentials['access_token']}"
                        ),
                    }
                    return self._user_credentials
        except ClientResponseError as error:
            if error.status in [HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN]:
                raise UnauthorizedError from error
            raise TractiveError from error
        except Exception as error:
            raise TractiveError from error

        return None

    async def close(self) -> None:
        """Close the session."""
        if self.session and self._close_session:
            await self.session.close()

    def base_headers(self) -> dict[str, str]:
        """Get base headers."""
        return {
            "x-tractive-client": self._client_id,
            "content-type": "application/json;charset=UTF-8",
            "accept": "application/json, text/plain, */*",
        }
